@@ -379,9 +379,14 @@ type EUnary struct {
 }
 
 type EBinary struct {
-	Op    OpCode
 	Left  Expr
 	Right Expr
+	Op    OpCode
+
+	// If true, this expression was originally a self-assignment of the form
+	// "x = y". This is relevant because it's used to emit a warning later on
+	// about code that accidentally does "x = x".
+	WasIdentifierAssign bool
 }
 
 type EBoolean struct{ Value bool }
@@ -503,6 +508,10 @@ type EIdentifier struct {
 	// unwrapped if the resulting value is unused. Unwrapping means discarding
 	// the call target but keeping any arguments with side effects.
 	CallCanBeUnwrappedIfUnused bool
+
+	// If true, this identifier was originally a "x as T" or "<T>x" expression.
+	// This is used to avoid warnings about "x = x as T" assignments.
+	WasTypeScriptCast bool
 }
 
 // This is similar to an EIdentifier but it represents a reference to an ES6
@@ -649,11 +658,11 @@ func (*ERequireResolve) isExpr()    {}
 func (*EImport) isExpr()            {}
 
 func Assign(a Expr, b Expr) Expr {
-	return Expr{a.Loc, &EBinary{BinOpAssign, a, b}}
+	return Expr{Loc: a.Loc, Data: &EBinary{Op: BinOpAssign, Left: a, Right: b}}
 }
 
 func AssignStmt(a Expr, b Expr) Stmt {
-	return Stmt{a.Loc, &SExpr{Expr{a.Loc, &EBinary{BinOpAssign, a, b}}}}
+	return Stmt{Loc: a.Loc, Data: &SExpr{Value: Assign(a, b)}}
 }
 
 func Not(a Expr) Expr {
@@ -661,7 +670,7 @@ func Not(a Expr) Expr {
 	if not, ok := a.Data.(*EUnary); ok && not.Op == UnOpNot && IsBooleanValue(not.Value) {
 		return not.Value
 	}
-	return Expr{a.Loc, &EUnary{UnOpNot, a}}
+	return Expr{Loc: a.Loc, Data: &EUnary{UnOpNot, a}}
 }
 
 func IsBooleanValue(a Expr) bool {
@@ -686,7 +695,7 @@ func IsBooleanValue(a Expr) bool {
 }
 
 func JoinWithComma(a Expr, b Expr) Expr {
-	return Expr{a.Loc, &EBinary{BinOpComma, a, b}}
+	return Expr{Loc: a.Loc, Data: &EBinary{Op: BinOpComma, Left: a, Right: b}}
 }
 
 func JoinAllWithComma(all []Expr) Expr {
@@ -748,8 +757,12 @@ type SExportDefault struct {
 }
 
 type ExportStarAlias struct {
-	Loc  logger.Loc
-	Name string
+	Loc logger.Loc
+
+	// Although this alias name starts off as being the same as the statement's
+	// namespace symbol, it may diverge if the namespace symbol name is minified.
+	// The original alias name is preserved here to avoid this scenario.
+	OriginalName string
 }
 
 type SExportStar struct {
@@ -771,6 +784,11 @@ type SLazyExport struct {
 
 type SExpr struct {
 	Value Expr
+
+	// This is set to true for automatically-generated expressions that should
+	// not affect tree shaking. For example, calling a function from the runtime
+	// that doesn't have externally-visible side effects.
+	DoesNotAffectTreeShaking bool
 }
 
 type EnumValue struct {
@@ -927,8 +945,8 @@ type SLocal struct {
 	IsExport bool
 
 	// The TypeScript compiler doesn't generate code for "import foo = bar"
-	// statements inside namespaces where the import is never used.
-	WasTSImportEqualsInNamespace bool
+	// statements where the import is never used.
+	WasTSImportEquals bool
 }
 
 type SBreak struct {
@@ -989,9 +1007,14 @@ type ClauseItem struct {
 	AliasLoc logger.Loc
 	Name     LocRef
 
-	// This is needed for "export {foo as bar} from 'path'" statements. This case
-	// is a re-export and "foo" and "bar" are both aliases. We need to preserve
-	// both aliases in case the symbol is renamed.
+	// This is the original name of the symbol stored in "Name". It's needed for
+	// "SExportClause" statements such as this:
+	//
+	//   export {foo as bar} from 'path'
+	//
+	// In this case both "foo" and "bar" are aliases because it's a re-export.
+	// We need to preserve both aliases in case the symbol is renamed. In this
+	// example, "foo" is "OriginalName" and "bar" is "Alias".
 	OriginalName string
 }
 
@@ -1077,6 +1100,9 @@ const (
 	// the module. Presumably this is because the imports may be type-only.
 	SymbolImport
 
+	// Assigning to a "const" symbol will throw a TypeError at runtime
+	SymbolConst
+
 	// This annotates all other symbols that don't have special behavior.
 	SymbolOther
 
@@ -1357,7 +1383,7 @@ type AST struct {
 	// since we already have to traverse the AST then anyway and the parser pass
 	// is conveniently fully parallelized.
 	NamedImports            map[Ref]NamedImport
-	NamedExports            map[string]Ref
+	NamedExports            map[string]NamedExport
 	TopLevelSymbolToParts   map[Ref][]uint32
 	ExportStarImportRecords []uint32
 
@@ -1463,7 +1489,7 @@ func (minifier *NameMinifier) NumberToMinifiedName(i int) string {
 }
 
 func (ast *AST) HasCommonJSFeatures() bool {
-	return ast.HasTopLevelReturn || ast.UsesExportsRef || ast.UsesModuleRef
+	return ast.UsesCommonJSExports() || ast.HasTopLevelReturn
 }
 
 func (ast *AST) UsesCommonJSExports() bool {
@@ -1488,6 +1514,11 @@ type NamedImport struct {
 	IsExported bool
 }
 
+type NamedExport struct {
+	Ref      Ref
+	AliasLoc logger.Loc
+}
+
 // Each file is made up of multiple parts, and each part consists of one or
 // more top-level statements. Parts are used for tree shaking and code
 // splitting analysis. Individual parts of a file can be discarded by tree
@@ -1519,7 +1550,7 @@ type Part struct {
 	CanBeRemovedIfUnused bool
 
 	// If true, this is the automatically-generated part for this file's ES6
-	// exports. It may hold the "const exports = {};" statement and also the
+	// exports. It may hold the "var exports = {};" statement and also the
 	// "__export(exports, { ... })" call to initialize the getters.
 	IsNamespaceExport bool
 
@@ -1601,7 +1632,7 @@ func MergeSymbols(symbols SymbolMap, old Ref, new Ref) Ref {
 // This has a custom implementation instead of using "filepath.Dir/Base/Ext"
 // because it should work the same on Unix and Windows. These names end up in
 // the generated output and the generated output should not depend on the OS.
-func platformIndependentPathDirBaseExt(path string) (dir string, base string, ext string) {
+func PlatformIndependentPathDirBaseExt(path string) (dir string, base string, ext string) {
 	for {
 		i := strings.LastIndexAny(path, "/\\")
 
@@ -1642,20 +1673,22 @@ func platformIndependentPathDirBaseExt(path string) (dir string, base string, ex
 // collisions.
 func GenerateNonUniqueNameFromPath(path string) string {
 	// Get the file name without the extension
-	dir, base, _ := platformIndependentPathDirBaseExt(path)
+	dir, base, _ := PlatformIndependentPathDirBaseExt(path)
 
 	// If the name is "index", use the directory name instead. This is because
 	// many packages in npm use the file name "index.js" because it triggers
 	// node's implicit module resolution rules that allows you to import it by
 	// just naming the directory.
 	if base == "index" {
-		_, dirBase, _ := platformIndependentPathDirBaseExt(dir)
+		_, dirBase, _ := PlatformIndependentPathDirBaseExt(dir)
 		if dirBase != "" {
 			base = dirBase
 		}
 	}
 
-	// Convert it to an ASCII identifier
+	// Convert it to an ASCII identifier. Note: If you change this to a non-ASCII
+	// identifier, you're going to potentially cause trouble with non-BMP code
+	// points in target environments that don't support bracketed Unicode escapes.
 	bytes := []byte{}
 	needsGap := false
 	for _, c := range base {
