@@ -16,11 +16,11 @@ func expectParseErrorTS(t *testing.T, contents string, expected string) {
 	t.Run(contents, func(t *testing.T) {
 		t.Helper()
 		log := logger.NewDeferLog()
-		Parse(log, test.SourceForTest(contents), config.Options{
+		Parse(log, test.SourceForTest(contents), OptionsFromConfig(&config.Options{
 			TS: config.TSOptions{
 				Parse: true,
 			},
-		})
+		}))
 		msgs := log.Done()
 		text := ""
 		for _, msg := range msgs {
@@ -35,11 +35,11 @@ func expectPrintedTS(t *testing.T, contents string, expected string) {
 	t.Run(contents, func(t *testing.T) {
 		t.Helper()
 		log := logger.NewDeferLog()
-		tree, ok := Parse(log, test.SourceForTest(contents), config.Options{
+		tree, ok := Parse(log, test.SourceForTest(contents), OptionsFromConfig(&config.Options{
 			TS: config.TSOptions{
 				Parse: true,
 			},
-		})
+		}))
 		msgs := log.Done()
 		text := ""
 		for _, msg := range msgs {
@@ -52,7 +52,7 @@ func expectPrintedTS(t *testing.T, contents string, expected string) {
 		symbols := js_ast.NewSymbolMap(1)
 		symbols.Outer[0] = tree.Symbols
 		r := renamer.NewNoOpRenamer(symbols)
-		js := js_printer.Print(tree, symbols, r, js_printer.PrintOptions{}).JS
+		js := js_printer.Print(tree, symbols, r, js_printer.Options{}).JS
 		test.AssertEqual(t, string(js), expected)
 	})
 }
@@ -62,14 +62,14 @@ func expectParseErrorTSX(t *testing.T, contents string, expected string) {
 	t.Run(contents, func(t *testing.T) {
 		t.Helper()
 		log := logger.NewDeferLog()
-		Parse(log, test.SourceForTest(contents), config.Options{
+		Parse(log, test.SourceForTest(contents), OptionsFromConfig(&config.Options{
 			TS: config.TSOptions{
 				Parse: true,
 			},
 			JSX: config.JSXOptions{
 				Parse: true,
 			},
-		})
+		}))
 		msgs := log.Done()
 		text := ""
 		for _, msg := range msgs {
@@ -84,14 +84,14 @@ func expectPrintedTSX(t *testing.T, contents string, expected string) {
 	t.Run(contents, func(t *testing.T) {
 		t.Helper()
 		log := logger.NewDeferLog()
-		tree, ok := Parse(log, test.SourceForTest(contents), config.Options{
+		tree, ok := Parse(log, test.SourceForTest(contents), OptionsFromConfig(&config.Options{
 			TS: config.TSOptions{
 				Parse: true,
 			},
 			JSX: config.JSXOptions{
 				Parse: true,
 			},
-		})
+		}))
 		msgs := log.Done()
 		text := ""
 		for _, msg := range msgs {
@@ -104,7 +104,7 @@ func expectPrintedTSX(t *testing.T, contents string, expected string) {
 		symbols := js_ast.NewSymbolMap(1)
 		symbols.Outer[0] = tree.Symbols
 		r := renamer.NewNoOpRenamer(symbols)
-		js := js_printer.Print(tree, symbols, r, js_printer.PrintOptions{}).JS
+		js := js_printer.Print(tree, symbols, r, js_printer.Options{}).JS
 		test.AssertEqual(t, string(js), expected)
 	})
 }
@@ -112,6 +112,7 @@ func expectPrintedTSX(t *testing.T, contents string, expected string) {
 func TestTSTypes(t *testing.T) {
 	expectPrintedTS(t, "let x: T extends number\n ? T\n : number", "let x;\n")
 	expectPrintedTS(t, "let x: {y: T extends number ? T : number}", "let x;\n")
+	expectPrintedTS(t, "let x: {y: T \n extends: number}", "let x;\n")
 	expectPrintedTS(t, "let x: {y: T \n extends?: number}", "let x;\n")
 	expectPrintedTS(t, "let x: (number | string)[]", "let x;\n")
 	expectPrintedTS(t, "let x: [string[]?]", "let x;\n")
@@ -148,6 +149,8 @@ func TestTSTypes(t *testing.T) {
 	expectPrintedTS(t, "type Foo<T> = {readonly [P in keyof T]: T[P]}", "")
 	expectPrintedTS(t, "type Foo<T> = {-readonly [P in keyof T]: T[P]}", "")
 	expectPrintedTS(t, "type Foo<T> = {+readonly [P in keyof T]: T[P]}", "")
+	expectPrintedTS(t, "let x: number! = y", "let x = y;\n")
+	expectPrintedTS(t, "let x: number \n !y", "let x;\n!y;\n")
 	expectPrintedTS(t, "const x: unique symbol = y", "const x = y;\n")
 	expectPrintedTS(t, "let x: typeof a = y", "let x = y;\n")
 	expectPrintedTS(t, "let x: typeof a.b = y", "let x = y;\n")
@@ -173,6 +176,15 @@ func TestTSTypes(t *testing.T) {
 	expectPrintedTS(t, "let foo = bar\nas (null)", "let foo = bar;\nas(null);\n")
 	expectParseErrorTS(t, "let foo = (bar\nas (null))", "<stdin>: error: Expected \")\" but found \"as\"\n")
 
+	expectPrintedTS(t, "a as any ? b : c;", "a ? b : c;\n")
+	expectPrintedTS(t, "a as any ? async () => b : c;", "a ? async () => b : c;\n")
+	expectPrintedTS(t, "foo as number extends Object ? any : any;", "foo;\n")
+	expectPrintedTS(t, "foo as number extends Object ? () => void : any;", "foo;\n")
+	expectPrintedTS(t, "let a = b ? c : d as T extends T ? T extends T ? T : never : never ? e : f;", "let a = b ? c : d ? e : f;\n")
+	expectParseErrorTS(t, "type a = b extends c", "<stdin>: error: Expected \"?\" but found end of file\n")
+	expectParseErrorTS(t, "type a = b extends c extends d", "<stdin>: error: Expected \"?\" but found \"extends\"\n")
+	expectParseErrorTS(t, "type a = b ? c : d", "<stdin>: error: Expected \";\" but found \"?\"\n")
+
 	expectPrintedTS(t, "let foo: keyof Object = 'toString'", "let foo = \"toString\";\n")
 	expectPrintedTS(t, "let foo: keyof\nObject = 'toString'", "let foo = \"toString\";\n")
 	expectPrintedTS(t, "let foo: (keyof\nObject) = 'toString'", "let foo = \"toString\";\n")
@@ -184,6 +196,36 @@ func TestTSTypes(t *testing.T) {
 	expectPrintedTS(t, "let foo: `${'a' | 'b'}-${'c' | 'd'}` = 'a-c'", "let foo = \"a-c\";\n")
 }
 
+func TestTSAsCast(t *testing.T) {
+	expectPrintedTS(t, "x as any\n(y);", "x;\ny;\n")
+	expectPrintedTS(t, "x as any\n`y`;", "x;\n`y`;\n")
+	expectPrintedTS(t, "x as any\n`${y}`;", "x;\n`${y}`;\n")
+	expectPrintedTS(t, "x as any\n--y;", "x;\n--y;\n")
+	expectPrintedTS(t, "x as any\n++y;", "x;\n++y;\n")
+	expectPrintedTS(t, "x + y as any\n(z as any) + 1;", "x + y;\nz + 1;\n")
+	expectPrintedTS(t, "x + y as any\n(z as any) = 1;", "x + y;\nz = 1;\n")
+	expectPrintedTS(t, "x = y as any\n(z as any) + 1;", "x = y;\nz + 1;\n")
+	expectPrintedTS(t, "x = y as any\n(z as any) = 1;", "x = y;\nz = 1;\n")
+	expectPrintedTS(t, "x * y as any\n['z'];", "x * y;\n[\"z\"];\n")
+	expectPrintedTS(t, "x * y as any\n.z;", "x * y;\n")
+	expectPrintedTS(t, "x as y['x'];", "x;\n")
+	expectPrintedTS(t, "x as y!['x'];", "x;\n")
+	expectPrintedTS(t, "x as y\n['x'];", "x;\n[\"x\"];\n")
+	expectPrintedTS(t, "x as y\n!['x'];", "x;\n![\"x\"];\n")
+	expectParseErrorTS(t, "x = y as any `z`;", "<stdin>: error: Expected \";\" but found \"`z`\"\n")
+	expectParseErrorTS(t, "x = y as any `${z}`;", "<stdin>: error: Expected \";\" but found \"`${\"\n")
+	expectParseErrorTS(t, "x = y as any?.z;", "<stdin>: error: Expected \";\" but found \"?.\"\n")
+	expectParseErrorTS(t, "x = y as any--;", "<stdin>: error: Expected \";\" but found \"--\"\n")
+	expectParseErrorTS(t, "x = y as any++;", "<stdin>: error: Expected \";\" but found \"++\"\n")
+	expectParseErrorTS(t, "x = y as any(z);", "<stdin>: error: Expected \";\" but found \"(\"\n")
+	expectParseErrorTS(t, "x = y as any\n= z;", "<stdin>: error: Unexpected \"=\"\n")
+	expectParseErrorTS(t, "a, x as y `z`;", "<stdin>: error: Expected \";\" but found \"`z`\"\n")
+	expectParseErrorTS(t, "a ? b : x as y `z`;", "<stdin>: error: Expected \";\" but found \"`z`\"\n")
+	expectParseErrorTS(t, "x as any = y;", "<stdin>: error: Expected \";\" but found \"=\"\n")
+	expectParseErrorTS(t, "(x as any = y);", "<stdin>: error: Expected \")\" but found \"=\"\n")
+	expectParseErrorTS(t, "(x = y as any(z));", "<stdin>: error: Expected \")\" but found \"(\"\n")
+}
+
 func TestTSClass(t *testing.T) {
 	expectPrintedTS(t, "export default class Foo {}", "export default class Foo {\n}\n")
 	expectPrintedTS(t, "export default class Foo extends Bar<T> {}", "export default class Foo extends Bar {\n}\n")
@@ -351,13 +393,17 @@ export let x;
 })(x || (x = {}));
 `)
 
+	errorText := `<stdin>: error: "foo" has already been declared
+<stdin>: note: "foo" was originally declared here
+`
+
 	// Namespaces with values are not allowed to merge
-	expectParseErrorTS(t, "var foo; namespace foo { 0 }", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "let foo; namespace foo { 0 }", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "const foo = 0; namespace foo { 0 }", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } var foo", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } let foo", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } const foo = 0", "<stdin>: error: \"foo\" has already been declared\n")
+	expectParseErrorTS(t, "var foo; namespace foo { 0 }", errorText)
+	expectParseErrorTS(t, "let foo; namespace foo { 0 }", errorText)
+	expectParseErrorTS(t, "const foo = 0; namespace foo { 0 }", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } var foo", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } let foo", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } const foo = 0", errorText)
 
 	// Namespaces without values are allowed to merge
 	expectPrintedTS(t, "var foo; namespace foo {}", "var foo;\n")
@@ -413,10 +459,10 @@ export let x;
   0;
 })(foo || (foo = {}));
 `)
-	expectParseErrorTS(t, "namespace foo { 0 } function foo() {}", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } function* foo() {}", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } async function foo() {}", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } class foo {}", "<stdin>: error: \"foo\" has already been declared\n")
+	expectParseErrorTS(t, "namespace foo { 0 } function foo() {}", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } function* foo() {}", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } async function foo() {}", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } class foo {}", errorText)
 	expectPrintedTS(t, "namespace foo { 0 } enum foo { a }", `var foo;
 (function(foo) {
   0;
@@ -464,11 +510,11 @@ async function foo() {
 `)
 
 	// Namespace merging shouldn't allow for other merging
-	expectParseErrorTS(t, "class foo {} namespace foo { 0 } class foo {}", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "class foo {} namespace foo { 0 } enum foo {}", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "enum foo {} namespace foo { 0 } class foo {}", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } namespace foo { 0 } let foo", "<stdin>: error: \"foo\" has already been declared\n")
-	expectParseErrorTS(t, "namespace foo { 0 } enum foo {} class foo {}", "<stdin>: error: \"foo\" has already been declared\n")
+	expectParseErrorTS(t, "class foo {} namespace foo { 0 } class foo {}", errorText)
+	expectParseErrorTS(t, "class foo {} namespace foo { 0 } enum foo {}", errorText)
+	expectParseErrorTS(t, "enum foo {} namespace foo { 0 } class foo {}", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } namespace foo { 0 } let foo", errorText)
+	expectParseErrorTS(t, "namespace foo { 0 } enum foo {} class foo {}", errorText)
 
 	// Test dot nested namespace syntax
 	expectPrintedTS(t, "namespace foo.bar { foo(bar) }", `var foo;
@@ -650,15 +696,12 @@ func TestTSNamespaceExports(t *testing.T) {
 		namespace A {
 			export namespace B {
 				export const foo = 1
-				foo += foo
 			}
 			namespace C {
 				export const foo = 1
-				foo += foo
 			}
 			namespace D {
 				const foo = 1
-				foo += foo
 			}
 		}
 	`, `var A;
@@ -666,17 +709,14 @@ func TestTSNamespaceExports(t *testing.T) {
   let B;
   (function(B) {
     B.foo = 1;
-    B.foo += B.foo;
   })(B = A.B || (A.B = {}));
   let C;
   (function(C) {
     C.foo = 1;
-    C.foo += C.foo;
   })(C || (C = {}));
   let D;
   (function(D) {
     const foo = 1;
-    foo += foo;
   })(D || (D = {}));
 })(A || (A = {}));
 `)
@@ -749,6 +789,50 @@ func TestTSNamespaceExports(t *testing.T) {
   console.log(E);
   console.log(N);
 })(ns || (ns = {}));
+`)
+
+	expectPrintedTS(t, `
+		namespace a { export var a = 123; log(a) }
+		namespace b { export let b = 123; log(b) }
+		namespace c { export enum c {} log(c) }
+		namespace d { export class d {} log(d) }
+		namespace e { export namespace e {} log(e) }
+		namespace f { export function f() {} log(f) }
+	`, `var a;
+(function(_a) {
+  _a.a = 123;
+  log(_a.a);
+})(a || (a = {}));
+var b;
+(function(_b) {
+  _b.b = 123;
+  log(_b.b);
+})(b || (b = {}));
+var c;
+(function(_c) {
+  let c;
+  (function(c) {
+  })(c = _c.c || (_c.c = {}));
+  log(c);
+})(c || (c = {}));
+var d;
+(function(_d) {
+  class d {
+  }
+  _d.d = d;
+  log(d);
+})(d || (d = {}));
+var e;
+(function(e) {
+  log(e);
+})(e || (e = {}));
+var f;
+(function(_f) {
+  function f() {
+  }
+  _f.f = f;
+  log(f);
+})(f || (f = {}));
 `)
 }
 
@@ -1008,7 +1092,37 @@ func TestTSDeclare(t *testing.T) {
 	expectPrintedTS(t, "declare module 'X'\n{ let foo }", "")
 	expectPrintedTS(t, "declare global { interface Foo {} let foo: any } let bar", "let bar;\n")
 	expectPrintedTS(t, "declare module M { const x }", "")
-	expectParseErrorTS(t, "module M { const x }", "<stdin>: error: This constant must be initialized\n")
+	expectPrintedTS(t, "declare module M { global { const x } }", "")
+	expectPrintedTS(t, "declare module M { global { const x } function foo() {} }", "")
+	expectPrintedTS(t, "declare module M { global \n { const x } }", "")
+	expectPrintedTS(t, "declare module M { import 'path' }", "")
+	expectPrintedTS(t, "declare module M { import x from 'path' }", "")
+	expectPrintedTS(t, "declare module M { import {x} from 'path' }", "")
+	expectPrintedTS(t, "declare module M { import * as ns from 'path' }", "")
+	expectPrintedTS(t, "declare module M { import foo = bar }", "")
+	expectPrintedTS(t, "declare module M { export import foo = bar }", "")
+	expectPrintedTS(t, "declare module M { export {x} from 'path' }", "")
+	expectPrintedTS(t, "declare module M { export default 123 }", "")
+	expectPrintedTS(t, "declare module M { export default function x() {} }", "")
+	expectPrintedTS(t, "declare module M { export default class X {} }", "")
+	expectPrintedTS(t, "declare module M { export * as ns from 'path' }", "")
+	expectPrintedTS(t, "declare module M { export * from 'path' }", "")
+	expectPrintedTS(t, "declare module M { export = foo }", "")
+	expectPrintedTS(t, "declare module M { export as namespace ns }", "")
+	expectPrintedTS(t, "declare module M { export as namespace ns; }", "")
+	expectParseErrorTS(t, "declare module M { export as namespace ns.foo }", "<stdin>: error: Expected \";\" but found \".\"\n")
+	expectParseErrorTS(t, "declare module M { export as namespace ns function foo() {} }", "<stdin>: error: Expected \";\" but found \"function\"\n")
+	expectParseErrorTS(t, "module M { const x }", "<stdin>: error: The constant \"x\" must be initialized\n")
+	expectParseErrorTS(t, "module M { const [] }", "<stdin>: error: This constant must be initialized\n")
+	expectParseErrorTS(t, "module M { const {} }", "<stdin>: error: This constant must be initialized\n")
+
+	// This is a weird case where "," after a rest parameter is allowed
+	expectPrintedTS(t, "declare function fn(x: any, ...y, )", "")
+	expectPrintedTS(t, "declare function fn(x: any, ...y: any, )", "")
+	expectParseErrorTS(t, "function fn(x: any, ...y, )", "<stdin>: error: Expected \")\" but found \",\"\n")
+	expectParseErrorTS(t, "function fn(x: any, ...y, ) {}", "<stdin>: error: Expected \")\" but found \",\"\n")
+	expectParseErrorTS(t, "function fn(x: any, ...y: any, )", "<stdin>: error: Expected \")\" but found \",\"\n")
+	expectParseErrorTS(t, "function fn(x: any, ...y: any, ) {}", "<stdin>: error: Expected \")\" but found \",\"\n")
 }
 
 func TestTSDecorator(t *testing.T) {
@@ -1062,6 +1176,24 @@ func TestTSDecorator(t *testing.T) {
 	expectParseErrorTS(t, "class Foo { @dec static async* #foo() {} }", "<stdin>: error: Expected identifier but found \"#foo\"\n")
 }
 
+func TestTSTry(t *testing.T) {
+	expectPrintedTS(t, "try {} catch (x: any) {}", "try {\n} catch (x) {\n}\n")
+	expectPrintedTS(t, "try {} catch (x: unknown) {}", "try {\n} catch (x) {\n}\n")
+	expectPrintedTS(t, "try {} catch (x: number) {}", "try {\n} catch (x) {\n}\n")
+
+	expectPrintedTS(t, "try {} catch ({x}: any) {}", "try {\n} catch ({x}) {\n}\n")
+	expectPrintedTS(t, "try {} catch ({x}: unknown) {}", "try {\n} catch ({x}) {\n}\n")
+	expectPrintedTS(t, "try {} catch ({x}: number) {}", "try {\n} catch ({x}) {\n}\n")
+
+	expectPrintedTS(t, "try {} catch ([x]: any) {}", "try {\n} catch ([x]) {\n}\n")
+	expectPrintedTS(t, "try {} catch ([x]: unknown) {}", "try {\n} catch ([x]) {\n}\n")
+	expectPrintedTS(t, "try {} catch ([x]: number) {}", "try {\n} catch ([x]) {\n}\n")
+
+	expectParseErrorTS(t, "try {} catch (x!) {}", "<stdin>: error: Expected \")\" but found \"!\"\n")
+	expectParseErrorTS(t, "try {} catch (x!: any) {}", "<stdin>: error: Expected \")\" but found \"!\"\n")
+	expectParseErrorTS(t, "try {} catch (x!: unknown) {}", "<stdin>: error: Expected \")\" but found \"!\"\n")
+}
+
 func TestTSArrow(t *testing.T) {
 	expectPrintedTS(t, "(a?) => {}", "(a) => {\n};\n")
 	expectPrintedTS(t, "(a?: number) => {}", "(a) => {\n};\n")
@@ -1163,21 +1295,32 @@ func TestTSImport(t *testing.T) {
 func TestTSImportEquals(t *testing.T) {
 	expectPrintedTS(t, "import x = require('foo'); x()", "const x = require(\"foo\");\nx();\n")
 	expectPrintedTS(t, "import x = require('foo')\nx()", "const x = require(\"foo\");\nx();\n")
-	expectPrintedTS(t, "import x = foo.bar; x()", "var x = foo.bar;\nx();\n")
-	expectPrintedTS(t, "import x = foo.bar\nx()", "var x = foo.bar;\nx();\n")
+	expectPrintedTS(t, "import x = require\nx()", "const x = require;\nx();\n")
+	expectPrintedTS(t, "import x = foo.bar; x()", "const x = foo.bar;\nx();\n")
+	expectPrintedTS(t, "import x = foo.bar\nx()", "const x = foo.bar;\nx();\n")
+	expectParseErrorTS(t, "import x = foo()", "<stdin>: error: Expected \";\" but found \"(\"\n")
+	expectParseErrorTS(t, "import x = foo<T>.bar", "<stdin>: error: Expected \";\" but found \"<\"\n")
 	expectParseErrorTS(t, "{ import x = foo.bar }", "<stdin>: error: Unexpected \"x\"\n")
 
 	expectPrintedTS(t, "export import x = require('foo'); x()", "export const x = require(\"foo\");\nx();\n")
 	expectPrintedTS(t, "export import x = require('foo')\nx()", "export const x = require(\"foo\");\nx();\n")
-	expectPrintedTS(t, "export import x = foo.bar; x()", "export var x = foo.bar;\nx();\n")
-	expectPrintedTS(t, "export import x = foo.bar\nx()", "export var x = foo.bar;\nx();\n")
+	expectPrintedTS(t, "export import x = foo.bar; x()", "export const x = foo.bar;\nx();\n")
+	expectPrintedTS(t, "export import x = foo.bar\nx()", "export const x = foo.bar;\nx();\n")
 
 	expectParseError(t, "export import foo = bar", "<stdin>: error: Unexpected \"import\"\n")
 	expectParseErrorTS(t, "export import {foo} from 'bar'", "<stdin>: error: Expected identifier but found \"{\"\n")
 	expectParseErrorTS(t, "export import foo from 'bar'", "<stdin>: error: Expected \"=\" but found \"from\"\n")
 	expectParseErrorTS(t, "export import foo = bar; var x; export {x as foo}",
-		"<stdin>: error: Multiple exports with the same name \"foo\"\n")
+		`<stdin>: error: Multiple exports with the same name "foo"
+<stdin>: note: "foo" was originally exported here
+`)
 	expectParseErrorTS(t, "{ export import foo = bar }", "<stdin>: error: Unexpected \"export\"\n")
+
+	errorText := `<stdin>: warning: This assignment will throw because "x" is a constant
+<stdin>: note: "x" was declared a constant here
+`
+	expectParseErrorTS(t, "import x = require('y'); x = require('z')", errorText)
+	expectParseErrorTS(t, "import x = y.z; x = z.y", errorText)
 }
 
 func TestTSImportEqualsInNamespace(t *testing.T) {
@@ -1185,7 +1328,7 @@ func TestTSImportEqualsInNamespace(t *testing.T) {
 	expectPrintedTS(t, "namespace ns { import foo = bar; type x = foo.x }", "")
 	expectPrintedTS(t, "namespace ns { import foo = bar.x; foo }", `var ns;
 (function(ns) {
-  var foo = bar.x;
+  const foo = bar.x;
   foo;
 })(ns || (ns = {}));
 `)
@@ -1216,9 +1359,16 @@ func TestTSTypeOnlyImport(t *testing.T) {
 	expectPrintedTS(t, "import type {foo, bar as baz} from 'bar'; x", "x;\n")
 	expectPrintedTS(t, "import type {foo, bar as baz} from 'bar'\nx", "x;\n")
 
-	expectPrintedTS(t, "import type = bar", "var type = bar;\n")
+	expectPrintedTS(t, "import type = bar; type", "const type = bar;\ntype;\n")
+	expectPrintedTS(t, "import type = foo.bar; type", "const type = foo.bar;\ntype;\n")
+	expectPrintedTS(t, "import type = require('type'); type", "const type = require(\"type\");\ntype;\n")
 	expectPrintedTS(t, "import type from 'bar'; type", "import type from \"bar\";\ntype;\n")
 
+	expectPrintedTS(t, "import a = b; import c = a.c", "")
+	expectPrintedTS(t, "import c = a.c; import a = b", "")
+	expectPrintedTS(t, "import a = b; import c = a.c; c()", "const a = b;\nconst c = a.c;\nc();\n")
+	expectPrintedTS(t, "import c = a.c; import a = b; c()", "const c = a.c;\nconst a = b;\nc();\n")
+
 	expectParseErrorTS(t, "import type", "<stdin>: error: Expected \"from\" but found end of file\n")
 	expectParseErrorTS(t, "import type * foo", "<stdin>: error: Expected \"as\" but found \"foo\"\n")
 	expectParseErrorTS(t, "import type * as 'bar'", "<stdin>: error: Expected identifier but found \"'bar'\"\n")
@@ -1360,3 +1510,17 @@ h();
 Foo[_b] = 1;
 `)
 }
+
+func TestWarningReassign(t *testing.T) {
+	expectParseErrorTS(t, "let x; x = x", "<stdin>: warning: Assignment of \"x\" to itself has no effect\n")
+	expectParseErrorTS(t, "let x; x = (x)", "<stdin>: warning: Assignment of \"x\" to itself has no effect\n")
+	expectParseErrorTS(t, "let x; x = (true ? x : x)", "")
+	expectParseErrorTS(t, "let x; x = (false ? x : x)", "")
+	expectParseErrorTS(t, "let x; x = (0, x)", "")
+	expectParseErrorTS(t, "let x; (x as any) = x", "")
+	expectParseErrorTS(t, "let x; x = x as any", "")
+	expectParseErrorTS(t, "let x; (<any>x) = x", "")
+	expectParseErrorTS(t, "let x; x = <any>x", "")
+	expectParseErrorTS(t, "let x; x! = x", "")
+	expectParseErrorTS(t, "let x; x = x!", "")
+}
